#include "jp/ggaf/dx/actor/GeometricActor.h"

#include "jp/ggaf/dx/God.h"
#include "jp/ggaf/core/util/Status.h"
#include "jp/ggaf/dx/actor/supporter/VecDriver.h"
#include "jp/ggaf/dx/actor/supporter/GeoDriver.h"
#include "jp/ggaf/dx/actor/supporter/Scaler.h"
#include "jp/ggaf/dx/actor/supporter/SeTransmitterForActor.h"
#include "jp/ggaf/dx/actor/supporter/Checker.h"
#include "jp/ggaf/dx/scene/Spacetime.h"
#include "jp/ggaf/dx/util/Util.h"

using namespace GgafDx;

GeometricActor::GeometricActor(const char* prm_name,
                               Checker* prm_pChecker) : BaseActor(prm_name),
_pVecDriver(nullptr),
_pGeoDriver(nullptr),
_pScaler(nullptr),
_pSeTransmitter(nullptr),
_pDriverLeader(nullptr),
_is_2D(false),
_offscreen_kind(-1),
_x(0), _y(0), _z(0),
_rx(0), _ry(0), _rz(0),
_sx(R_SC(1.0)), _sy(R_SC(1.0)), _sz(R_SC(1.0)),
_pChecker(prm_pChecker),
_rate_of_bounding_sphere_radius(1.0f),
_fX(C_DX(_x)), _fY(C_DX(_y)), _fZ(C_DX(_z)),
_dest_from_vppln_top(0.0f),
_dest_from_vppln_bottom(0.0f),
_dest_from_vppln_left(0.0f),
_dest_from_vppln_right(0.0f),
_dest_from_vppln_infront(0.0f),
_dest_from_vppln_back(0.0f),
_pFunc_calc_rot_mv_world_matrix(nullptr),
_matWorld(),
_matWorldRotMv(),
_matInvWorldRotMv(),
_was_calculated_matInvWorldRotMv(false),
_pActor_base(nullptr),
_x_local(_x), _y_local(_y), _z_local(_z),
_rx_local(_rx), _ry_local(_ry), _rz_local(_rz),
_x_final(_x), _y_final(_y), _z_final(_z),
_rx_final(_rx), _ry_final(_ry), _rz_final(_rz),
_is_local(false)
{
    _obj_class |= Obj_GgafDx_GeometricActor;
    _class_name = "GeometricActor";
    _pFormation = nullptr;
}
VecDriver* GeometricActor::getVecDriver() {
    return _pVecDriver ? _pVecDriver : _pVecDriver = NEW VecDriver(this);
}

GeoDriver* GeometricActor::getGeoDriver() {
    return _pGeoDriver ? _pGeoDriver : _pGeoDriver = NEW GeoDriver(this);
}

DriverLeader* GeometricActor::createCurveDriverLeader(CurveManufacture* prm_pCurveManufacture) {
    DriverLeader* pDriverLeader = nullptr;
    CurveManufacture::MoveDriver move_driver = prm_pCurveManufacture->_move_driver;
    if (move_driver == CurveManufacture::MoveDriver::GeoDriver) {
        pDriverLeader = prm_pCurveManufacture->createGeoDriverLeader(getGeoDriver());
    } else if (move_driver == CurveManufacture::MoveDriver::VecDriver) {
        pDriverLeader = prm_pCurveManufacture->createVecDriverLeader(getVecDriver());
    } else {
        throwCriticalException("createCurveDriverLeader() : CurveManufacture::MoveDrive ݒ肳Ă܂B"<<
                "ldr_file="<<prm_pCurveManufacture->_ldr_file<<" move_driver="<<move_driver<<" this="<<NODE_INFO<<"");
    }
    return pDriverLeader;
}


SeTransmitterForActor* GeometricActor::getSeTransmitter() {
    return _pSeTransmitter ? _pSeTransmitter : _pSeTransmitter = NEW SeTransmitterForActor(this);
}

Scaler* GeometricActor::getScaler() {
    return _pScaler ? _pScaler : _pScaler = NEW Scaler(this);
}

void GeometricActor::processSettlementBehavior() {
    if (_is_2D) {
        // 
    } else {

        _was_calculated_matInvWorldRotMv = false; //ts񖢌vZɃZbg

        if (_pActor_base) {
            //y䂠莞[JWɈU߂
            changeGeoLocal();
        }

        //DirectX̒PʂɍWϊĂiWorldϊs쐬ɂgp܂j
        _fX = C_DX(_x);
        _fY = C_DX(_y);
        _fZ = C_DX(_z);
        //Worldϊsi_matWorldjXV
        if (_pFunc_calc_rot_mv_world_matrix) {
            //]~ړ̂݌vZ _matWorldRotMv 
            (*_pFunc_calc_rot_mv_world_matrix)(this, _matWorldRotMv);
            _matWorld = _matWorldRotMv; //Rs[ĕێ
            //]~ړ ̑O XP[lāA
            //ŏII _matWorld  s(g~]~ړ)ێ
            if (_sx != LEN_UNIT) {
                const float sx = SC_R(_sx);
                _matWorld._11 *= sx;
                _matWorld._12 *= sx;
                _matWorld._13 *= sx;
            }
            if (_sy != LEN_UNIT) {
                const float sy = SC_R(_sy);
                _matWorld._21 *= sy;
                _matWorld._22 *= sy;
                _matWorld._23 *= sy;
            }
            if (_sz != LEN_UNIT) {
                const float sz = SC_R(_sz);
                _matWorld._31 *= sz;
                _matWorld._32 *= sz;
                _matWorld._33 *= sz;
            }
        }

        //ftHgł́A_matWorldRotMv = ]ϊs ~ sړϊs
        //                _matWorld      = gkϊs ~ _matWorldRotMv ƂȂ悤ɂĂB
        //܂ _matWorld = gk]sړ
        //_matWorldRotMv  appendGroupChildAsFk() sɎgp邽߂ɍ쐬ĂB
        //] appendGroupChildAsFk() ΎgpȂȂ΁A_matWorldRotMv̌vZ͕svB
        //processSettlementBehavior() I[o[ChA
        //ϊs쐬ƒP邱ƂŁAœK\B

        if (_pActor_base) {
            //΍Wɕϊ
            D3DXMatrixMultiply(&_matWorld     , &_matWorld     , &(_pActor_base->_matWorldRotMv)); //
            D3DXMatrixMultiply(&_matWorldRotMv, &_matWorldRotMv, &(_pActor_base->_matWorldRotMv)); //
            changeGeoFinal();
            //[hϊs~y䃏[hϊśu]~ړv݂̂̐ρj畽sړoŏIIȍWƂ
            _fX = _matWorld._41;
            _fY = _matWorld._42;
            _fZ = _matWorld._43;
            _x = DX_C(_fX);
            _y = DX_C(_fY);
            _z = DX_C(_fZ);

            //[Jłfaceς݂̃[hϊśu]~ړv݂̂̐(_matWorldRotMv)̂ŁA
            //xNg(1,0,0)ɍςݕϊs|΍ŏIIface𓾂
            UTIL::convVectorToRzRy(_matWorldRotMv._11, _matWorldRotMv._12, _matWorldRotMv._13, _rz, _ry);
            _rx = _rx_local; //̂܂
            //Ă̂B
            //xNg̓[hϊs̐ρi_matWorldRotMv)ŕϊA݂̍ŏIIȌɌB
            //匳̕xNg(x_org_,y_org_,z_org_)A
            //[hϊs̉]̐ρi_matWorldRotMv)̐ mat_xxA
            //ŏIIȕxNg(vX, vY, vZ) Ƃ
            //
            //                         | mat_11 mat_12 mat_13 |
            //| x_org_ y_org_ z_org_ | | mat_21 mat_22 mat_23 | = | vX vY vZ |
            //                         | mat_31 mat_32 mat_33 |
            //ƂȂB
            //
            //vX = x_org_*mat_11 + y_org_*mat_21 + z_org_*mat_31
            //vY = x_org_*mat_12 + y_org_*mat_22 + z_org_*mat_32
            //vZ = x_org_*mat_13 + y_org_*mat_23 + z_org_*mat_33
            //
            //ĂŁA匳O̒PʕxNg(1,0,0)̏ꍇ͂ǂȂ邩Hl
            //
            //vX = x_org_*mat_11
            //vY = x_org_*mat_12
            //vZ = x_org_*mat_13
            //
            //ƂȂB{Avł́Af͑S(1,0,0)OƂĂ邽
            //ŏIIȕxNǵix_org_*mat_11, x_org_*mat_12, x_org_*mat_13) ƂȂB
            //̕xNg _rz _ry \ΗǂB
            //vZ₷悤x_org_1ƒu
            //
            //UTIL::convVectorToRzRy(_matWorldRotMv._11, _matWorldRotMv._12, _matWorldRotMv._13, _rz, _ry);
            //ƂȂ
            processChangeGeoFinal(); //΍WvZ̏pR[obN
        }

        //ʂ̋XV
        const dxcoord fX = _fX;
        const dxcoord fY = _fY;
        const dxcoord fZ = _fZ;
        static Camera* pCam = pGOD->getSpacetime()->getCamera();
        static const D3DXPLANE* pPlnTop = &(pCam->_plnTop);
        _dest_from_vppln_top     = pPlnTop->a * fX +
                                   pPlnTop->b * fY +
                                   pPlnTop->c * fZ +
                                   pPlnTop->d;
        static const D3DXPLANE* pPlnBottom = &(pCam->_plnBottom);
        _dest_from_vppln_bottom  = pPlnBottom->a * fX +
                                   pPlnBottom->b * fY +
                                   pPlnBottom->c * fZ +
                                   pPlnBottom->d;
        static const D3DXPLANE* pPlnLeft = &(pCam->_plnLeft);
        _dest_from_vppln_left    = pPlnLeft->a * fX +
                                   pPlnLeft->b * fY +
                                   pPlnLeft->c * fZ +
                                   pPlnLeft->d;
        static const D3DXPLANE* pPlnRight = &(pCam->_plnRight);
        _dest_from_vppln_right   = pPlnRight->a * fX +
                                   pPlnRight->b * fY +
                                   pPlnRight->c * fZ +
                                   pPlnRight->d;
        static const D3DXPLANE* pPlnInfront = &(pCam->_plnInfront);
        _dest_from_vppln_infront = pPlnInfront->a * fX +
                                   pPlnInfront->b * fY +
                                   pPlnInfront->c * fZ +
                                   pPlnInfront->d;
        static const D3DXPLANE* pPlnBack = &(pCam->_plnBack);
        _dest_from_vppln_back    = pPlnBack->a * fX +
                                   pPlnBack->b * fY +
                                   pPlnBack->c * fZ +
                                   pPlnBack->d;
        _offscreen_kind = -1;
    }
}

void GeometricActor::processPreJudgement() {
    if (_pChecker) {
        if (_can_hit_flg) {
            if (_can_hit_out_of_view == false && isOutOfView()) {
                //O蔻薳̏ꍇ͓o^Ȃ
            } else  {
                //ؓo^
                _pChecker->updateHitArea();
            }
        }
    }
}

void GeometricActor::changeGeoLocal() {
#ifdef MY_DEBUG
        if (!_pActor_base) {
            throwCriticalException("changeGeoLocal() : yAN^[܂BmFĉBthis="<<NODE_INFO<<"");
        }
        if (_is_local) {
            throwCriticalException("changeGeoLocal() : Ƀ[JWnłBΉmFĉBthis="<<NODE_INFO<<"");
        }
#endif
        _x_final  = _x;
        _y_final  = _y;
        _z_final  = _z;
        _rx_final = _rx;
        _ry_final = _ry;
        _rz_final = _rz;
        _x  = _x_local;
        _y  = _y_local;
        _z  = _z_local;
        _rx = _rx_local;
        _ry = _ry_local;
        _rz = _rz_local;
        _is_local = true;
    }

void GeometricActor::changeGeoFinal() {
#ifdef MY_DEBUG
    if (!_pActor_base) {
        throwCriticalException("changeGeoFinal() : yAN^[܂BmFĉBthis="<<NODE_INFO<<"");
    }
    if (!_is_local) {
        throwCriticalException("changeGeoFinal() : Ƀ[JWnłBΉmFĉBthis="<<NODE_INFO<<"");
    }
#endif
    _x_local = _x;
    _y_local = _y;
    _z_local = _z;
    _rx_local = _rx;
    _ry_local = _ry;
    _rz_local = _rz;
    _x  = _x_final;
    _y  = _y_final;
    _z  = _z_final;
    _rx = _rx_final;
    _ry = _ry_final;
    _rz = _rz_final;
    _is_local = false;
}

//void GeometricActor::updateGeoFinalFromLocal() {
//
//}
GgafCore::GroupHead* GeometricActor::appendGroupChildAsFk(kind_t prm_kind,
                                                          GeometricActor* prm_pGeoActor,
                                                          coord prm_x_init_local,
                                                          coord prm_y_init_local,
                                                          coord prm_z_init_local,
                                                          coord prm_rx_init_local,
                                                          coord prm_ry_init_local,
                                                          coord prm_rz_init_local) {
#ifdef MY_DEBUG
    if (_pFunc_calc_rot_mv_world_matrix) {
        //OK
    } else {
        throwCriticalException(": "<<
                                   "this="<<NODE_INFO<<" ́A_pFunc_calc_rot_mv_world_matrix  nullptrׁ̈AFKx[XƂȂ鎑i܂");
    }
#endif
    GgafCore::GroupHead* pGroupHead = appendGroupChild(prm_kind, prm_pGeoActor);
    prm_pGeoActor->_pActor_base = this;
    prm_pGeoActor->changeGeoLocal();
    prm_pGeoActor->_x = prm_x_init_local;
    prm_pGeoActor->_y = prm_y_init_local;
    prm_pGeoActor->_z = prm_z_init_local;
    prm_pGeoActor->_rx = prm_rx_init_local;
    prm_pGeoActor->_ry = prm_ry_init_local;
    prm_pGeoActor->_rz = prm_rz_init_local;
//    prm_pGeoActor->getVecDriver()->_rz_mv = prm_rz_init_local;
//    prm_pGeoActor->getVecDriver()->_ry_mv = prm_ry_init_local;

    prm_pGeoActor->changeGeoFinal();
    return pGroupHead;
}

GgafCore::GroupHead* GeometricActor::appendGroupChildAsFk(GeometricActor* prm_pGeoActor,
                                                          coord prm_x_init_local,
                                                          coord prm_y_init_local,
                                                          coord prm_z_init_local,
                                                          coord prm_rx_init_local,
                                                          coord prm_ry_init_local,
                                                          coord prm_rz_init_local) {
    return appendGroupChildAsFk(prm_pGeoActor->getDefaultKind(),
                                prm_pGeoActor,
                                prm_x_init_local,
                                prm_y_init_local,
                                prm_z_init_local,
                                prm_rx_init_local,
                                prm_ry_init_local,
                                prm_rz_init_local);
}

bool GeometricActor::processHitChkLogic(GgafCore::Actor* prm_pOtherActor) {
    if (_can_hit_flg && prm_pOtherActor->_can_hit_flg) {
        //&& prm_pOtherActor->instanceOf(Obj_GgafDx_GeometricActor)) { 蔻肪̂GeometricActorȏƔf
        //_can_hit_flg && prm_pOtherActor->_can_hit_flg ̃`FbN͔ؓo^Oɂ`FbNĂ邪
        //łx`FbNقǂB
        //ȂȂ΁Aʂȃqbg`FbNsȂ߁AonHit(GgafCore::Actor*)  setHitAble(false) sA
        //Qdqbg`FbNh~sĂ邩ȂB
        if (_pChecker) {
            return _pChecker->isHit(((GeometricActor*)prm_pOtherActor)->_pChecker);
        }
    }
    return false;
}

int GeometricActor::isOutOfView() {
    //Eaif̂ĵŁAPX_DX(1) ʓɂ邩ǂ𔻒肷B
    //fʓɂ邩ǂ́AFigureActor ŃI[o[ChĂōsB
    if (_offscreen_kind == -1) {
        const dxcoord bound = PX_DX(1); //Ea
        if (_dest_from_vppln_top < bound) {
            if (_dest_from_vppln_bottom < bound) {
                if (_dest_from_vppln_left < bound) {
                    if (_dest_from_vppln_right < bound) {
                        if (_dest_from_vppln_infront < bound) {
                            if (_dest_from_vppln_back < bound) {
                                //Viewport͈͓
                                _offscreen_kind = 0;
                            } else {
                                //ʂ艜Ŕ͈͊O
                                _offscreen_kind = 6;
                            }
                        } else {
                            //OʂOŔ͈͊O
                            _offscreen_kind = 5;
                        }
                    } else {
                        //EʂEŔ͈͊O
                        _offscreen_kind = 4;
                    }
                } else {
                    //ʂ荶Ŕ͈͊O
                    _offscreen_kind = 3;
                }
            } else {
                //ʂ艺Ŕ͈͊O
                _offscreen_kind = 2;
            }
        } else {
            //㕽ʂŔ͈͊O
            _offscreen_kind = 1;
        }
    }
    return _offscreen_kind;
}

bool GeometricActor::isOutOfSpacetime() const {
    Spacetime* pSpacetime =  pGOD->getSpacetime();
    if (pSpacetime->_x_bound_left <= _x) {
        if (_x <= pSpacetime->_x_bound_right) {
            if (pSpacetime->_y_bound_bottom <= _y) {
                if (_y <= pSpacetime->_y_bound_top) {
                    if (pSpacetime->_z_bound_near <= _z) {
                        if (_z <= pSpacetime->_z_bound_far) {
                            return false;
                        }
                    }
                }
            }
        }
    }
    return true;
}

void GeometricActor::defineRotMvWorldMatrix(void (*prm_pFunc)(const GeometricActor*, D3DXMATRIX&)) {
    _pFunc_calc_rot_mv_world_matrix = prm_pFunc;
    (*_pFunc_calc_rot_mv_world_matrix)(this, _matWorldRotMv);
}

void GeometricActor::defineRotMvWorldMatrix_Billboard() {
    defineRotMvWorldMatrix(Util::setWorldMatrix_RzBxyzMv);
}

void GeometricActor::setPositionAt(const GeoElem* prm_pGeoElem) {
    _x = prm_pGeoElem->x;
    _y = prm_pGeoElem->y;
    _z = prm_pGeoElem->z;
}

void GeometricActor::setPositionAround(coord x, coord y, coord z, coord r) {
    double out_x, out_y, out_z;
    UTIL::getRndSpherCoord(r, out_x, out_y, out_z);
    setPosition(x+out_x, y+out_y, z+out_z);
}

void GeometricActor::setPositionAround(coord x, coord y, coord r) {
    double out_x, out_y;
    UTIL::getRndCircleCoord(r, out_x, out_y);
    setPosition(x+out_x, y+out_y);
}

void GeometricActor::setPositionAroundAt(const GeometricActor* prm_pActor, coord r) {
    setPositionAround(prm_pActor->_x, prm_pActor->_y, prm_pActor->_z, r);
}
void GeometricActor::setPositionAroundAt(const GeoElem* prm_pGeoElem, coord r) {
    setPositionAround(prm_pGeoElem->x, prm_pGeoElem->y, prm_pGeoElem->z, r);
}
void GeometricActor::setFaceAngAs(const GeometricActor* prm_pActor) {
    setRxFaceAng(prm_pActor->_rx);
    setRyFaceAng(prm_pActor->_ry);
    setRzFaceAng(prm_pActor->_rz);
}
void GeometricActor::setFaceAngAs(const GeoElem* prm_pGeoElem) {
    setRxFaceAng(prm_pGeoElem->rx);
    setRyFaceAng(prm_pGeoElem->ry);
    setRzFaceAng(prm_pGeoElem->rz);
}

//void GeometricActor::setFaceAng(axis prm_axis, angle prm_face) {
//    _face[prm_axis] = UTIL::simplifyAng(prm_face);
//}

void GeometricActor::setRzFaceAng(angle prm_face) {
    _rz = UTIL::simplifyAng(prm_face);
}
void GeometricActor::setRyFaceAng(angle prm_face) {
    _ry = UTIL::simplifyAng(prm_face);
}
void GeometricActor::setRxFaceAng(angle prm_face) {
    _rx = UTIL::simplifyAng(prm_face);
}

void GeometricActor::setRzRyFaceAng(angle prm_rz_face, angle prm_ry_face) {
    setRzFaceAng(prm_rz_face);
    setRyFaceAng(prm_ry_face);
}

void GeometricActor::setFaceAngTwd(coord prm_tx, coord prm_ty, coord prm_tz) {
    coord vx = prm_tx - _x;
    coord vy = prm_ty - _y;
    coord vz = prm_tz - _z;
    if (vx == 0 && vy == 0 && vz == 0) {
        //g̍Wɓ̂ŁAȂ
    } else {
        UTIL::convVectorToRzRy(vx, vy, vz,
                               _rz, _ry );
    }
}
void GeometricActor::setFaceAngTwd(const GeometricActor* prm_pActor) {
    setFaceAngTwd(prm_pActor->_x, prm_pActor->_y, prm_pActor->_z);
}

void GeometricActor::scaleAs(const GeometricActor* prm_pActor) {
    _sx = prm_pActor->_sx;
    _sy = prm_pActor->_sy;
    _sz = prm_pActor->_sz;
}

void GeometricActor::onEnd() {
    BaseActor::onEnd();
    _pFormation = nullptr;
}

GeometricActor::~GeometricActor() {
    GGAF_DELETE_NULLABLE(_pVecDriver);
    GGAF_DELETE_NULLABLE(_pGeoDriver);
    GGAF_DELETE_NULLABLE(_pScaler);
    GGAF_DELETE_NULLABLE(_pSeTransmitter);
}

void GeometricActor::dump() {
    _TRACE_("\t\t\t\t\t\t\t\t"<<NODE_INFO<<"("<<_x<<","<<_y<<","<<_z<<")"<<DUMP_FLGS);
    GgafCore::Actor* pActor_tmp = _pChildFirst;
    while (pActor_tmp) {
        pActor_tmp->dump("\t\t\t\t\t\t\t\tb");
        if (pActor_tmp->getNext()) {
            pActor_tmp = pActor_tmp->getNext();
        } else {
            _TRACE_("yxz"<<NODE_INFO<<" nextnullptrĂ܂");
            break;
        }
        if (pActor_tmp->isFirst()) {
            _TRACE_("\t\t\t\t\t\t\t\t");
            break;
        }
    }
}

void GeometricActor::dump(std::string prm_parent) {
    _TRACE_(prm_parent <<NODE_INFO<<"("<<_x<<","<<_y<<","<<_z<<")"<<DUMP_FLGS);
    GgafCore::Actor* pActor_tmp = _pChildFirst;
    while (pActor_tmp) {
        pActor_tmp->dump(prm_parent + "b");
        if (pActor_tmp->getNext()) {
            pActor_tmp = pActor_tmp->getNext();
        } else {
            _TRACE_("yxz"<<NODE_INFO<<" nextnullptrĂ܂");
            break;
        }
        if (pActor_tmp->isFirst()) {
            _TRACE_(prm_parent+"");
            break;
        }
    }
}

