#include "jp/ggaf/lib/actor/laserchip/RefractionLaserChip.h"

#include "jp/ggaf/dx/actor/supporter/VecDriver.h"
#include "jp/ggaf/lib/actor/laserchip/LaserChipDepository.h"



using namespace GgafLib;

RefractionLaserChip::RefractionLaserChip(const char* prm_name, const char* prm_model) :
        LaserChip(prm_name, prm_model) {
    _class_name = "RefractionLaserChip";
    _is_leader = false;
    _num_refraction = 1;
    _frame_standstill_refraction = 10;
    _frame_between_refraction = 20;

    _cnt_refraction = 0;
    _frame_refraction_enter = 0;
    _frame_refraction_out = 0;
    _is_refracting = false;

    _pDepo_refraction_effect = nullptr;
    _pRefractionEffect = nullptr;
    _prev_pRefractionEffect = nullptr;

    _begining_x = _x;
    _begining_y = _y;
    _begining_z = _z;
    _begining_rx = _rx;
    _begining_ry = _ry;
    _begining_rz = _rz;
    _begining_rz_mv = getVecDriver()->_rz_mv;
    _begining_ry_mv = getVecDriver()->_ry_mv;
    _begining_velo_mv   = getVecDriver()->_velo_mv;
    _prev_x  = _x;
    _prev_y  = _y;
    _prev_z  = _z;
    _prev_rx = _rx;
    _prev_ry = _ry;
    _prev_rz = _rz;
    _prev_rz_mv = getVecDriver()->_rz_mv;
    _prev_ry_mv = getVecDriver()->_ry_mv;
    _prev_velo_mv   = getVecDriver()->_velo_mv;
    _prev_is_refracting = false;
    _is_fix_begin_pos = true;
    _refraction_end_frames = 0;
}

void RefractionLaserChip::config(int prm_num_refraction,
                                 frame prm_frame_between_refraction,
                                 frame prm_frame_standstill_refraction,
                                 bool prm_is_fix_begin_pos,
                                 GgafCore::ActorDepository* prm_pDepo_refraction_effect) {
#ifdef MY_DEBUG
    if (prm_frame_between_refraction == 0) {
        throwCriticalException("iԊut[0͐ݒsłBname="<<getName());
    }
    if (prm_frame_standstill_refraction == 0) {
        throwCriticalException("ܗt[0͐ݒsłBname="<<getName());
    }
#endif
    _num_refraction = prm_num_refraction;
    _frame_between_refraction = prm_frame_between_refraction;
    _frame_standstill_refraction = prm_frame_standstill_refraction;
    _is_fix_begin_pos = prm_is_fix_begin_pos;
    _pDepo_refraction_effect = prm_pDepo_refraction_effect;
    _refraction_end_frames = (_num_refraction * _frame_standstill_refraction) +
                             ((_num_refraction+1) * prm_frame_between_refraction);
}

void RefractionLaserChip::onActive() {
    //Ǝݒ肵ꍇApĕʃNX쐬AI[o[ChĂB
    //̍ ́A{NX onActive() \bhĂяoĂB
    LaserChip::onActive();
    GgafDx::VecDriver* pVecDriver = getVecDriver();
    RefractionLaserChip* pChip_infront =  (RefractionLaserChip*)_pChip_infront;
    //[U[`bvo
    if (pChip_infront == nullptr) {
        _is_leader = true;
        //g擪̏ꍇ
        _begining_x = _x;
        _begining_y = _y;
        _begining_z = _z;
        _begining_rx = _rx;
        _begining_ry = _ry;
        _begining_rz = _rz;
        _begining_rz_mv = pVecDriver->_rz_mv;
        _begining_ry_mv = pVecDriver->_ry_mv;
        _begining_velo_mv   = pVecDriver->_velo_mv;
        _cnt_refraction = 0;
        _frame_refraction_enter = getBehaveingFrame() + _frame_between_refraction + 1;
        _frame_refraction_out = _frame_refraction_enter + _frame_standstill_refraction;
        //onRefractionInto(_cnt_refraction);
        onRefractionOutOf(_cnt_refraction);  //R[obN 0ڂ̋܏IX^[g
    } else {
        _is_leader = false;
        _begining_x = pChip_infront->_begining_x;
        _begining_y = pChip_infront->_begining_y;
        _begining_z = pChip_infront->_begining_z;
        _begining_rx = pChip_infront->_begining_rx;
        _begining_ry = pChip_infront->_begining_ry;
        _begining_rz = pChip_infront->_begining_rz;
        _begining_rz_mv = pChip_infront->_begining_rz_mv;
        _begining_ry_mv = pChip_infront->_begining_ry_mv;
        _begining_velo_mv   = pChip_infront->_begining_velo_mv;
        if (_is_fix_begin_pos) {
            _x = _begining_x;
            _y = _begining_y;
            _z = _begining_z;
            _rx = _begining_rx;
            _ry = _begining_ry;
            _rz = _begining_rz;
            pVecDriver->setRzRyMvAng(_begining_rz_mv, _begining_ry_mv);
            pVecDriver->setMvVelo(_begining_velo_mv);
        }
        _cnt_refraction = 0;
        _frame_refraction_enter = INT_MAX;
        _frame_refraction_out = INT_MAX;
        _refraction_end_frames = pChip_infront->_refraction_end_frames;
    }

    _is_refracting = false;
    _prev_is_refracting = false;
}

void RefractionLaserChip::onInactive() {
    //[U[`bv

    //̃`bv΁ÄӎviړȂǁjp
    //́A[U[ꂽꍇAȉ̃p[^[݈̂pAړp邽߁B
    //xAړ\ȂǈpȂ̂̂ŁAGȈړۂ́ARRɒӂ𕥂ƁI
    //[U[Q[̈OɂƂA擪`bv珇ɘAňp邱ƂɂȂB
    //ƖʂۂǁAȂΐ擪̎̃`bv̈OɌĈړƂ͌Ȃ̂ŁA͂KvB
    //yӁz
    //擪`bv~iړx0j inactive ƂȂꍇA~p邽߁A
    //RefractionLaser ͂ɗ܂荞ł܂B͉邱ƁB
    if (_pChip_behind) {
        RefractionLaserChip* const pChip_behind = (RefractionLaserChip*)_pChip_behind;
        GgafDx::VecDriver* const pChip_behind_pVecDriver = pChip_behind->getVecDriver();
        GgafDx::VecDriver* const pVecDriver = getVecDriver();
        pChip_behind->_rx = _rx;
        pChip_behind->_ry = _ry;
        pChip_behind->_rz = _rz;
        pChip_behind_pVecDriver->setRzRyMvAng(pVecDriver->_rz_mv, pVecDriver->_ry_mv);
        pChip_behind_pVecDriver->setMvVelo(pVecDriver->_velo_mv);
        pChip_behind->_cnt_refraction = _cnt_refraction;
        pChip_behind->_frame_refraction_enter = _frame_refraction_enter;
        pChip_behind->_frame_refraction_out = _frame_refraction_out;
        pChip_behind->_is_refracting = _is_refracting;
        //܃GtFNg
        if (_pRefractionEffect) {
            _pRefractionEffect->sayonara();
            _pRefractionEffect = nullptr;
        }
        if (_prev_pRefractionEffect) {
            _prev_pRefractionEffect->sayonara();
            _prev_pRefractionEffect = nullptr;
        }
        pChip_behind->_pRefractionEffect = _pRefractionEffect;
    } else {
        //܃GtFNg
        if (_pRefractionEffect) {
            _pRefractionEffect->sayonara();
            _pRefractionEffect = nullptr;
        }
        if (_prev_pRefractionEffect) {
            _prev_pRefractionEffect->sayonara();
            _prev_pRefractionEffect = nullptr;
        }
    }
    LaserChip::onInactive(); //Ȃؒf
}

void RefractionLaserChip::processBehavior() {
    //[U[`bv
    //Ǝݒ肵ꍇApĕʃNX쐬AI[o[ChĂB
    //̍ ́A{NX processBehavior() \bhĂяoĂB
    //Wɔf
    GgafDx::VecDriver* const pVecDriver = getVecDriver();
    RefractionLaserChip* pChip_infront =  (RefractionLaserChip*)_pChip_infront;
    if (getActiveFrame() > 1) { //Pt[ڂ́AݒWŕ\邽߁BړȂ
        //ActorDepository::dispatch() 
        //擾łꍇA|C^ԂƋɁÃAN^[̓AN^[҂̎q̈ԌɈړB
        //āA[U[̐擪珇ԂprocessBehavior() Ă΂邽߁Aȉ̂悤ɂ
        //qɂȂB
        if (pChip_infront == nullptr) {
            //{̐擪`bvA͂ɂ킩擪`bv̏ꍇ̋ʏ
            _prev_x  = _x;
            _prev_y  = _y;
            _prev_z  = _z;
            _prev_rx = _rx;
            _prev_ry = _ry;
            _prev_rz = _rz;
            _prev_rz_mv = pVecDriver->_rz_mv;
            _prev_ry_mv = pVecDriver->_ry_mv;
            _prev_velo_mv   = pVecDriver->_velo_mv;
            _prev_is_refracting = _is_refracting;
            _prev_pRefractionEffect = _pRefractionEffect;

            _pRefractionEffect = nullptr;
            if (!_is_refracting) {
                if (getBehaveingFrame() >= _frame_refraction_enter) {
                    if (_cnt_refraction < _num_refraction) {
                        _cnt_refraction++;
                        onRefractionInto(_cnt_refraction); //R[obN
                        _frame_refraction_out = getBehaveingFrame()  + _frame_standstill_refraction - 1; //0ōőBA]łPt
                        _is_refracting = true;

                        if (_pDepo_refraction_effect) {
                            _pRefractionEffect = (GgafDx::FigureActor*)_pDepo_refraction_effect->dispatch();
                            if (_pRefractionEffect) {
                                _pRefractionEffect->setPositionAt(this);
                                //ŒԂ̉\B
                                //̔qHiTODO:vjŃ[U[`bvłꍇAsayonara()oȂꍇB̏ꍇ̕یB
                                _pRefractionEffect->inactivateDelay(_pDepo->_num_chip_max +_frame_standstill_refraction);
                            }
                        }
                    }
                }
            }

            if (_is_refracting) {
                if (getBehaveingFrame() >= _frame_refraction_out) {
                    onRefractionOutOf(_cnt_refraction); //R[obN
                    _frame_refraction_enter = getBehaveingFrame() + _frame_between_refraction + 1;
                    //_frame_refraction_enter̔́At[łB
                    //getBehaveingFrame() ͎t[+1̂+1ĂKv
                    //Wς]
                    coord x = _x; coord y = _y; coord z = _z;
                    pVecDriver->behave(); //
                    _x = x; _y = y; _z = z;
                    _is_refracting = false;
                    return;
                }
            }

            if (!_is_refracting) {
                //_is_refracting͒~ȂĂ͂ȂgetVecDriver()->behave()sȂB
                //pVecDriver->behave();ȊOōW𑀍삵Ăꍇ́ASȒ~ɂȂȂ̂Œ
                pVecDriver->behave();
            }

        } else {
            //擪ȊÕ`bvq
            _prev_x  = _x;
            _prev_y  = _y;
            _prev_z  = _z;
            _prev_rx = _rx;
            _prev_ry = _ry;
            _prev_rz = _rz;
            _prev_rz_mv = pVecDriver->_rz_mv;
            _prev_ry_mv = pVecDriver->_ry_mv;
            _prev_velo_mv   = pVecDriver->_velo_mv;
            _prev_is_refracting = _is_refracting;
            _prev_pRefractionEffect = _pRefractionEffect;
            _x  = pChip_infront->_prev_x;
            _y  = pChip_infront->_prev_y;
            _z  = pChip_infront->_prev_z;
            _rx = pChip_infront->_prev_rx;
            _ry = pChip_infront->_prev_ry;
            _rz = pChip_infront->_prev_rz;
            pVecDriver->setRzRyMvAng(pChip_infront->_prev_rz_mv, pChip_infront->_prev_ry_mv);
            pVecDriver->setMvVelo(pChip_infront->_prev_velo_mv);
            _is_refracting =  pChip_infront->_prev_is_refracting;
            _pRefractionEffect = pChip_infront->_prev_pRefractionEffect;
            if (_pChip_behind == nullptr) {
                if (_pRefractionEffect) {
                    _pRefractionEffect->sayonara(_frame_standstill_refraction);
                }
            }
        }
    }
}

RefractionLaserChip::~RefractionLaserChip() {
}

